/*
 * Copyright (c) 2013 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * This file contains Original Code and/or Modifications of Original Code
 * as defined in and that are subject to the Apple Public Source License
 * Version 2.0 (the 'License'). You may not use this file except in
 * compliance with the License. Please obtain a copy of the License at
 * http://www.opensource.apple.com/apsl/ and read it before using this
 * file.
 * 
 * The Original Code and all software distributed under the License are
 * distributed on an 'AS IS' basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, QUIET ENJOYMENT OR NON-INFRINGEMENT.
 * Please see the License for the specific language governing rights and
 * limitations under the License.
 * 
 * @APPLE_LICENSE_HEADER_END@
 */

#include <stdio.h>
#include <asl.h>
#include <unistd.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <stdbool.h>
#include <errno.h>
#include <sys/dir.h>
#include <string.h>
#include <sys/sysctl.h>

/* Log files generated by RootDomain */
#define RD_STACKS "/var/log/SleepWakeStacks.dump"
#define RD_LOG    "/var/log/SleepWakeLog.dump"

/* Log files generated by RootDomain for AppleOSXWatchdog */
#define RD_APPLEOSXWATCHDOG_STACKS  "/var/log/AppleOSXWatchdogStacks.dump"
#define RD_APPLEOSXWATCHDOG_LOG     "/var/log/AppleOSXWatchdogLog.dump"

#define APPLEOSXWATCHDOG_FILENAMETOKEN "progress watchdog"
#define SLEEP_WAKE_FILENAMETOKEN "Sleep Wake Failure"

#define PANIC_FILE "/var/db/PanicReporter/current.panic"

void process_logfiles(char *stacksfile, char *logfile)
{
    struct stat stacks;
    struct stat logs;
    char cmd[200];
    DIR *dirp;
    struct dirent *dp;
    char fname[128], swfile[128];
    long crtime = 0;
    struct stat fstats, pfstat;
    struct timeval sleeptime, boottime;
    size_t size;
    bool isAppleOSXWatchdog = false;


    if ((lstat(logfile, &logs) != 0) || !S_ISREG(logs.st_mode))
        return;

    if (lstat(stacksfile, &stacks)) {
        // For cases where there is no stack file generated
        snprintf(cmd, sizeof(cmd), "/usr/bin/touch %s", stacksfile);
        system(cmd);
    }
    else if (!S_ISREG(S_IFREG)) {
        // If file exists, it should be a regular file with no links
        return;
    }
    
    /*
     * We need to differentiate Sleep/Wake failure from AppleOSXWatchdog failure. Indeed,
     * currently spindump has a different interface for both watchdogs.
     */
    isAppleOSXWatchdog = (stacksfile == RD_APPLEOSXWATCHDOG_STACKS || logfile == RD_APPLEOSXWATCHDOG_LOG);
    
    if (isAppleOSXWatchdog) {
        snprintf(cmd, sizeof(cmd),
                 "/usr/sbin/spindump -progress_watchdog_stackshot_file %s -progress_watchdog_data_file %s",
                 stacksfile, logfile);
    } else {
        snprintf(cmd, sizeof(cmd),
                 "/usr/sbin/spindump -sleepwakefailure_stackshot_file %s -sleepwakefailure_data_file %s",
                 stacksfile, logfile);
    }
    
    // Invoke spindump.
    system(cmd);


    unlink(stacksfile);
    unlink(logfile);


    // Due diligence check to make sure this file is not generated due to system sleep
    size = sizeof(sleeptime);
    if ((sysctlbyname("kern.sleeptime", &sleeptime, &size, NULL, 0) != 0) || (sleeptime.tv_sec != 0))
        return;

    // For RD_LOG files, trigger a pop-up dialog to report that
    // system has rebooted due to Sleep Wake failure.
    //
    // Look for most recently created Sleep Wake Failure file
    dirp = opendir("/Library/Logs/DiagnosticReports");
    if (dirp == NULL) {
        return ;
    }

    while ((dp = readdir(dirp)) != NULL) {
        char *filename      = isAppleOSXWatchdog ? APPLEOSXWATCHDOG_FILENAMETOKEN : SLEEP_WAKE_FILENAMETOKEN;
        size_t filename_len = isAppleOSXWatchdog ? sizeof(APPLEOSXWATCHDOG_FILENAMETOKEN) : sizeof(SLEEP_WAKE_FILENAMETOKEN);
        
        if ((strnstr(dp->d_name, filename, filename_len) == dp->d_name)
                    && (dp->d_type == DT_REG)) {
            snprintf(fname, sizeof(fname), "/Library/Logs/DiagnosticReports/%s", dp->d_name);
            stat(fname, &fstats);

            if (S_ISREG(fstats.st_mode) && (crtime < fstats.st_birthtime)) {
                crtime = fstats.st_birthtime;
                strncpy(swfile, fname, sizeof(swfile));
            }
        }
    }

    // Check for existence of SleepWakeFailure log file
    if (crtime == 0)
        return;     

    // Due diligence check to make sure this file is generated after system boot
    size = sizeof(boottime);
    if ((sysctlbyname("kern.boottime", &boottime, &size, NULL, 0) != 0) || (boottime.tv_sec > crtime))
        return;

    // If panic file link already exists, skip linking this Sleep Wake failure file
    if (lstat(PANIC_FILE, &pfstat) == 0)
        return;

    symlink(swfile, PANIC_FILE);

}

int main(int argc, char **argv)
{

    process_logfiles(RD_STACKS, RD_LOG);
    
    /* AppleOSXWatchdog */
    process_logfiles(RD_APPLEOSXWATCHDOG_STACKS, RD_APPLEOSXWATCHDOG_LOG);

    return 0;
}
